#ifndef PB161_EXAMPLE_H
#define PB161_EXAMPLE_H

#include <algorithm>
#include <cassert>
#include <cstdint>
#include <iostream>

class encoder
{
private:
  const unsigned bytes_in;
  const unsigned bytes_out;
  const unsigned modulo;

  virtual uint8_t code_byte(uint8_t) const = 0;
  virtual uint8_t code_pad() const = 0;

protected:
  encoder(unsigned _bytes_in, unsigned _bytes_out, unsigned _modulo)
    : bytes_in(_bytes_in), bytes_out(_bytes_out), modulo(_modulo)
  {
  }

public:
  void encode_stream(std::istream&, std::ostream&);

  template <typename IN_ITERATOR, typename OUT_ITERATOR>
  void encode(IN_ITERATOR in, const IN_ITERATOR& end, OUT_ITERATOR out)
  {
    do {
      uint64_t block_in = 0;
      uint8_t block_out[bytes_out];
      std::fill(block_out, block_out + bytes_out, 0);

      // nacte nejvyse i znaku ze vstupu
      unsigned in_i = 0;
      unsigned bytes_pad = 0;
      for (in_i = 0; (in_i < bytes_in) && (in != end); ++in_i)
        block_in = (block_in << 8) | static_cast<uint8_t>(*(in++));

      // zbytek doplni 0 a zapamatuje si, kolik doplnil
      bytes_pad = bytes_in - in_i;
      for (unsigned i = 0; i < bytes_pad; ++i) block_in <<= 8;

      const unsigned real_input = bytes_in - bytes_pad;
      const unsigned bytes_to_pad =
        bytes_out - ((real_input * bytes_out + bytes_in - 1) / bytes_in);

      unsigned out_i = 0;
      // korektne zpracovany vystup ponacita
      for (out_i = 0; out_i < bytes_out; ++out_i) {
        block_out[bytes_out - (out_i + 1)] = code_byte(block_in % modulo);
        block_in /= modulo;
      }
      // pozice odpovidajici doplnovanym datum predvyplni
      for (out_i = 0; out_i < bytes_to_pad; ++out_i) {
        block_out[bytes_out - (out_i + 1)] = code_pad();
      }

      for (unsigned i = 0; i < bytes_out; ++i) *(out++) = block_out[i];

    } while (in != end);
  }
};

struct base64 : public encoder
{
  base64() : encoder(3, 4, 64) {}
private:
  uint8_t code_byte(uint8_t byte) const
  {
    assert(byte < 64);
    const uint8_t spec_cases[] = {'+', '/'};
    // following function implements the base64 encoding table
    if (byte < 26) {
      return byte - 0 + 'A';
    } else if (byte < 52) {
      return byte - 26 + 'a';
    } else if (byte < 62) {
      return byte - 52 + '0';
    } else {
      return spec_cases[byte - 62];
    }
  }
  uint8_t code_pad() const { return '='; }
};

struct ascii85 : public encoder
{
  ascii85() : encoder(4, 5, 85) {}
private:
  uint8_t code_byte(uint8_t byte) const { return byte + 33; }
  uint8_t code_pad() const { return code_byte(0); }
};

#endif // PB161_EXAMPLE_H
